@@ -9,13 +9,15 @@ from __future__ import unicode_literals
 import typing
 
 from .base import Opener
+from .registry import registry
 
-if False:  # typing.TYPE_CHECKING
+if typing.TYPE_CHECKING:
     from typing import Text
     from .parse import ParseResult
-    from ..memoryfs import MemoryFS
+    from ..memoryfs import MemoryFS  # noqa: F401
 
 
+@registry.install
 class MemOpener(Opener):
     """`MemoryFS` opener.
     """
