@@ -6,18 +6,27 @@ from __future__ import absolute_import
 from __future__ import print_function
 from __future__ import unicode_literals
 
+import collections
 import contextlib
 import typing
 
-import six
 import pkg_resources
 
 from .base import Opener
 from .errors import UnsupportedProtocol, EntryPointError
 from .parse import parse_fs_url
 
-if False:  # typing.TYPE_CHECKING
-    from typing import Iterator, List, Text, Tuple, Union
+if typing.TYPE_CHECKING:
+    from typing import (
+        Callable,
+        Dict,
+        Iterator,
+        List,
+        Text,
+        Type,
+        Tuple,
+        Union,
+    )
     from ..base import FS
 
 
@@ -25,34 +34,62 @@ class Registry(object):
     """A registry for `Opener` instances.
     """
 
-    def __init__(self, default_opener="osfs"):
-        # type: (Text) -> None
+    def __init__(self, default_opener="osfs", load_extern=False):
+        # type: (Text, bool) -> None
         """Create a registry object.
 
         Arguments:
             default_opener (str, optional): The protocol to use, if one
                 is not supplied. The default is to use 'osfs', so that the
                 FS URL is treated as a system path if no protocol is given.
+            load_extern (bool, optional): Set to `True` to load openers from
+                PyFilesystem2 extensions. Defaults to `False`.
 
         """
         self.default_opener = default_opener
-        self._protocols = None  # type: Union[None, List[Text]]
+        self.load_extern = load_extern
+        self._protocols = {}  # type: Dict[Text, Opener]
 
     def __repr__(self):
         # type: () -> Text
         return "<fs-registry {!r}>".format(self.protocols)
 
+    def install(self, opener):
+        # type: (Union[Type[Opener], Opener, Callable[[], Opener]]) -> Opener
+        """Install an opener.
+
+        Arguments:
+            opener (`Opener`): an `Opener` instance, or a callable that
+                returns an opener instance.
+
+        Note:
+            May be used as a class decorator. For example::
+                registry = Registry()
+                @registry.install
+                class ArchiveOpener(Opener):
+                    protocols = ['zip', 'tar']
+        """
+        _opener = opener if isinstance(opener, Opener) else opener()
+        assert isinstance(_opener, Opener), "Opener instance required"
+        assert _opener.protocols, "must list one or more protocols"
+        for protocol in _opener.protocols:
+            self._protocols[protocol] = _opener
+        return _opener
+
     @property
     def protocols(self):
         # type: () -> List[Text]
         """`list`: the list of supported protocols.
         """
-        if self._protocols is None:
-            self._protocols = [
+
+        _protocols = list(self._protocols)
+        if self.load_extern:
+            _protocols.extend(
                 entry_point.name
                 for entry_point in pkg_resources.iter_entry_points("fs.opener")
-            ]
-        return self._protocols
+            )
+            _protocols = list(collections.OrderedDict.fromkeys(_protocols))
+        return _protocols
 
     def get_opener(self, protocol):
         # type: (Text) -> Opener
@@ -73,29 +110,41 @@ class Registry(object):
 
         """
         protocol = protocol or self.default_opener
-        entry_point = next(pkg_resources.iter_entry_points("fs.opener", protocol), None)
 
-        if entry_point is None:
-            raise UnsupportedProtocol("protocol '{}' is not supported".format(protocol))
-
-        try:
-            opener = entry_point.load()
-        except Exception as exception:
-            six.raise_from(
-                EntryPointError("could not load entry point; {}".format(exception)),
-                exception,
+        if self.load_extern:
+            entry_point = next(
+                pkg_resources.iter_entry_points("fs.opener", protocol), None
             )
         else:
+            entry_point = None
+
+        # If not entry point was loaded from the extensions, try looking
+        # into the registered protocols
+        if entry_point is None:
+            if protocol in self._protocols:
+                opener_instance = self._protocols[protocol]
+            else:
+                raise UnsupportedProtocol(
+                    "protocol '{}' is not supported".format(protocol)
+                )
+
+        # If an entry point was found in an extension, attempt to load it
+        else:
+            try:
+                opener = entry_point.load()
+            except Exception as exception:
+                raise EntryPointError(
+                    "could not load entry point; {}".format(exception)
+                )
             if not issubclass(opener, Opener):
                 raise EntryPointError("entry point did not return an opener")
 
-        try:
-            opener_instance = opener()
-        except Exception as exception:
-            six.raise_from(
-                EntryPointError("could not instantiate opener; {}".format(exception)),
-                exception,
-            )
+            try:
+                opener_instance = opener()
+            except Exception as exception:
+                raise EntryPointError(
+                    "could not instantiate opener; {}".format(exception)
+                )
 
         return opener_instance
 
@@ -140,7 +189,7 @@ class Registry(object):
 
     def open_fs(
         self,
-        fs_url,  # type: Text
+        fs_url,  # type: Union[FS, Text]
         writeable=False,  # type: bool
         create=False,  # type: bool
         cwd=".",  # type: Text
@@ -228,10 +277,8 @@ class Registry(object):
             _fs = self.open_fs(fs_url, create=create, writeable=writeable, cwd=cwd)
             try:
                 yield _fs
-            except:
-                raise
             finally:
                 _fs.close()
 
 
-registry = Registry()
+registry = Registry(load_extern=True)
