@@ -17,6 +17,7 @@ import platform
 import shutil
 import stat
 import sys
+import tempfile
 import typing
 
 import six
@@ -27,7 +28,7 @@ except ImportError:
     try:
         from scandir import scandir  # type: ignore
     except ImportError:  # pragma: no cover
-        scandir = None  # pragma: no cover
+        scandir = None  # type: ignore  # pragma: no cover
 
 try:
     from os import sendfile
@@ -35,10 +36,9 @@ except ImportError:
     try:
         from sendfile import sendfile  # type: ignore
     except ImportError:
-        sendfile = None  # pragma: no cover
+        sendfile = None  # type: ignore  # pragma: no cover
 
 from . import errors
-from .errors import FileExists
 from .base import FS
 from .enums import ResourceType
 from ._fscompat import fsencode, fsdecode, fspath
@@ -47,13 +47,13 @@ from .path import basename, dirname
 from .permissions import Permissions
 from .error_tools import convert_os_errors
 from .mode import Mode, validate_open_mode
-from .errors import NoURL
+from .errors import FileExpected, NoURL
+from ._url_tools import url_quote
 
-if False:  # typing.TYPE_CHECKING
+if typing.TYPE_CHECKING:
     from typing import (
         Any,
         BinaryIO,
-        Callable,
         Collection,
         Dict,
         Iterator,
@@ -90,6 +90,8 @@ class OSFS(FS):
         create_mode (int): The permissions that will be used to create
             the directory if ``create`` is `True` and the path doesn't
             exist, defaults to ``0o777``.
+        expand_vars(bool): If `True` (the default) environment variables of
+            the form $name or ${name} will be expanded.
 
     Raises:
         `fs.errors.CreateFailed`: If ``root_path`` does not
@@ -107,6 +109,7 @@ class OSFS(FS):
         root_path,  # type: Text
         create=False,  # type: bool
         create_mode=0o777,  # type: SupportsInt
+        expand_vars=True,  # type: bool
     ):
         # type: (...) -> None
         """Create an OSFS instance.
@@ -117,7 +120,9 @@ class OSFS(FS):
         self.root_path = root_path
         _drive, _root_path = os.path.splitdrive(fsdecode(fspath(root_path)))
         _root_path = _drive + (_root_path or "/") if _drive else _root_path
-        _root_path = os.path.expanduser(os.path.expandvars(_root_path))
+        _root_path = os.path.expanduser(
+            os.path.expandvars(_root_path) if expand_vars else _root_path
+        )
         _root_path = os.path.normpath(os.path.abspath(_root_path))
         self._root_path = _root_path
 
@@ -131,10 +136,10 @@ class OSFS(FS):
                 )
         else:
             if not os.path.isdir(_root_path):
-                raise errors.CreateFailed("root path does not exist")
+                message = "root path '{}' does not exist".format(_root_path)
+                raise errors.CreateFailed(message)
 
         _meta = self._meta = {
-            "case_insensitive": os.path.normcase("Aa") != "aa",
             "network": False,
             "read_only": False,
             "supports_rename": True,
@@ -143,6 +148,16 @@ class OSFS(FS):
             "virtual": False,
         }
 
+        try:
+            # https://stackoverflow.com/questions/7870041/check-if-file-system-is-case-insensitive-in-python
+            # I don't know of a better way of detecting case insensitivity of a
+            # filesystem
+            with tempfile.NamedTemporaryFile(prefix="TmP") as _tmp_file:
+                _meta["case_insensitive"] = os.path.exists(_tmp_file.name.lower())
+        except Exception:
+            if platform.system() != "Darwin":
+                _meta["case_insensitive"] = os.path.normcase("Aa") == "aa"
+
         if _WINDOWS_PLATFORM:  # pragma: no cover
             _meta["invalid_path_chars"] = (
                 "".join(six.unichr(n) for n in range(31)) + '\\:*?"<>|'
@@ -151,9 +166,13 @@ class OSFS(FS):
             _meta["invalid_path_chars"] = "\0"
 
             if "PC_PATH_MAX" in os.pathconf_names:
-                _meta["max_sys_path_length"] = os.pathconf(
-                    fsencode(_root_path), os.pathconf_names["PC_PATH_MAX"]
-                )
+                try:
+                    _meta["max_sys_path_length"] = os.pathconf(
+                        fsencode(_root_path), os.pathconf_names["PC_PATH_MAX"]
+                    )
+                except OSError:  # pragma: no cover
+                    # The above fails with nfs mounts on OSX. Go figure.
+                    pass
 
     def __repr__(self):
         # type: () -> str
@@ -168,7 +187,7 @@ class OSFS(FS):
         return fmt.format(_class_name.lower(), self.root_path)
 
     def _to_sys_path(self, path):
-        # type: (Text) -> Text
+        # type: (Text) -> bytes
         """Convert a FS path to a path on the OS.
         """
         sys_path = fsencode(
@@ -245,12 +264,15 @@ class OSFS(FS):
 
     def _gettarget(self, sys_path):
         # type: (Text) -> Optional[Text]
-        try:
-            target = os.readlink(fsencode(sys_path))
-        except OSError:
-            return None
-        else:
-            return target
+        if hasattr(os, "readlink"):
+            try:
+                if _WINDOWS_PLATFORM:  # pragma: no cover
+                    return os.readlink(sys_path)
+                else:
+                    return fsdecode(os.readlink(fsencode(sys_path)))
+            except OSError:
+                pass
+        return None
 
     def _make_link_info(self, sys_path):
         # type: (Text) -> Dict[Text, object]
@@ -328,6 +350,8 @@ class OSFS(FS):
         _mode.validate_bin()
         self.check()
         _path = self.validatepath(path)
+        if _path == "/":
+            raise errors.FileExpected(path)
         sys_path = self._to_sys_path(_path)
         with convert_os_errors("openbin", path):
             if six.PY2 and _mode.exclusive:
@@ -372,7 +396,7 @@ class OSFS(FS):
 
     # --- Type hint for opendir ------------------------------
 
-    if False:  # typing.TYPE_CHECKING
+    if typing.TYPE_CHECKING:
 
         def opendir(self, path, factory=None):
             # type: (_O, Text, Optional[_OpendirFactory]) -> SubFS[_O]
@@ -395,19 +419,20 @@ class OSFS(FS):
             raise errors.DirectoryExpected(dirname(dst_path))
         return _src_path, _dst_path
 
-    if sys.version_info[:2] < (3, 8) and sendfile is not None:
+    if sys.version_info[:2] <= (3, 8) and sendfile is not None:
 
-        _sendfile_error_codes = frozenset(
-            {
-                errno.EIO,
-                errno.EINVAL,
-                errno.ENOSYS,
-                errno.ENOTSUP,  # type: ignore
-                errno.EBADF,
-                errno.ENOTSOCK,
-                errno.EOPNOTSUPP,
-            }
-        )
+        _sendfile_error_codes = {
+            errno.EIO,
+            errno.EINVAL,
+            errno.ENOSYS,
+            errno.EBADF,
+            errno.ENOTSOCK,
+            errno.EOPNOTSUPP,
+        }
+
+        # PyPy doesn't define ENOTSUP so we have to add it conditionally.
+        if hasattr(errno, "ENOTSUP"):
+            _sendfile_error_codes.add(errno.ENOTSUP)
 
         def copy(self, src_path, dst_path, overwrite=False):
             # type: (Text, Text, bool) -> None
@@ -454,7 +479,12 @@ class OSFS(FS):
             self.check()
             namespaces = namespaces or ()
             _path = self.validatepath(path)
-            sys_path = self._to_sys_path(_path)
+            if _WINDOWS_PLATFORM:
+                sys_path = os.path.join(
+                    self._root_path, path.lstrip("/").replace("/", os.sep)
+                )
+            else:
+                sys_path = self._to_sys_path(_path)  # type: ignore
             with convert_os_errors("scandir", path, directory=True):
                 for dir_entry in scandir(sys_path):
                     info = {
@@ -498,7 +528,6 @@ class OSFS(FS):
             namespaces = namespaces or ()
             _path = self.validatepath(path)
             sys_path = self.getsyspath(_path)
-            _sys_path = fsencode(sys_path)
             with convert_os_errors("scandir", path, directory=True):
                 for entry_name in os.listdir(sys_path):
                     _entry_name = fsdecode(entry_name)
@@ -556,9 +585,14 @@ class OSFS(FS):
 
     def geturl(self, path, purpose="download"):
         # type: (Text, Text) -> Text
-        if purpose != "download":
+        sys_path = self.getsyspath(path)
+        if purpose == "download":
+            return "file://" + sys_path
+        elif purpose == "fs":
+            url_path = url_quote(sys_path)
+            return "osfs://" + url_path
+        else:
             raise NoURL(path, purpose)
-        return "file://" + self.getsyspath(path)
 
     def gettype(self, path):
         # type: (Text) -> ResourceType
@@ -595,6 +629,8 @@ class OSFS(FS):
         validate_open_mode(mode)
         self.check()
         _path = self.validatepath(path)
+        if _path == "/":
+            raise FileExpected(path)
         sys_path = self._to_sys_path(_path)
         with convert_os_errors("open", path):
             if six.PY2 and _mode.exclusive:
@@ -636,8 +672,7 @@ class OSFS(FS):
         except UnicodeEncodeError as error:
             raise errors.InvalidCharsInPath(
                 path,
-                msg="path '{path}' could not be encoded for the filesystem (check LANG env var); {error}".format(
-                    path=path, error=error
-                ),
+                msg="path '{path}' could not be encoded for the filesystem (check LANG"
+                    " env var); {error}".format(path=path, error=error),
             )
         return super(OSFS, self).validatepath(path)
